"""Module for miscellaneous functions and methods"""
import tempfile
import inspect
import platform
import os
from threading import Thread
import random
import string

from pyvista.utilities.errors import GPUInfo
import scooby
import pyvista
import numpy as np
import vtk

VTK9 = vtk.vtkVersion().GetVTKMajorVersion() >= 9


# path of this module
MODULE_PATH = os.path.dirname(inspect.getfile(inspect.currentframe()))


def get_ansys_bin(rver):
    """Identify the ansys executable based on the release version (e.g. "201")"""
    if os.name == 'nt':
        ans_root = 'c:/Program Files/ANSYS Inc/'
        mapdlbin = os.path.join(ans_root, 'v%s' % rver, 'ansys', 'bin', 'winx64',
                                'ANSYS%s.exe' % rver)
    else:
        ans_root = '/usr/ansys_inc'
        mapdlbin = os.path.join(ans_root, 'v%s' % rver, 'ansys', 'bin',
                                'ansys%s' % rver)

    return mapdlbin


def kill_process(proc_pid):
    """Kill a process with extreme prejudice"""
    import psutil  # imported here to avoid import errors when unused in windows
    process = psutil.Process(proc_pid)
    for proc in process.children(recursive=True):
        proc.kill()
    process.kill()


class Report(scooby.Report):
    """A class for custom scooby.Report."""

    def __init__(self, additional=None, ncol=3, text_width=80, sort=False,
                 gpu=True):
        """Generate a :class:`scooby.Report` instance.

        Parameters
        ----------
        additional : list(ModuleType), list(str)
            List of packages or package names to add to output information.

        ncol : int, optional
            Number of package-columns in html table; only has effect if
            ``mode='HTML'`` or ``mode='html'``. Defaults to 3.

        text_width : int, optional
            The text width for non-HTML display modes

        sort : bool, optional
            Alphabetically sort the packages

        gpu : bool
            Gather information about the GPU. Defaults to ``True`` but if
            experiencing renderinng issues, pass ``False`` to safely generate
            a report.

        """
        # Mandatory packages.
        core = ['pyvista', 'vtk', 'numpy', 'scipy', 'appdirs', 'pexpect',
                'ansys.mapdl.core', 'ansys.mapdl.reader']

        if os.name == 'linux':
            core.extend(['psutil', 'pexpect'])

        # Optional packages.
        optional = ['matplotlib', 'ansys.mapdl.corba']

        # Information about the GPU - bare except in case there is a rendering
        # bug that the user is trying to report.
        if gpu:
            try:
                extra_meta = [(t[1], t[0]) for t in GPUInfo().get_info()]
            except:
                extra_meta = ("GPU Details", "error")
        else:
            extra_meta = ("GPU Details", "None")

        scooby.Report.__init__(self, additional=additional, core=core,
                               optional=optional, ncol=ncol,
                               text_width=text_width, sort=sort,
                               extra_meta=extra_meta)

    def __repr__(self):
        add_text = '-'*79 + '\nPyMAPDL Software and Environment Report'
        return add_text + super().__repr__()


def is_float(input_string):
    """Returns true when a string can be converted to a float"""
    try:
        float(input_string)
        return True
    except ValueError:
        return False


def random_string(stringLength=10, letters=string.ascii_lowercase):
    """Generate a random string of fixed length """
    return ''.join(random.choice(letters) for i in range(stringLength))


def _configure_pyvista():
    """Configure PyVista's ``rcParams`` for pyansys"""
    import pyvista as pv
    pv.rcParams['interactive'] = True
    pv.rcParams["cmap"] = "jet"
    pv.rcParams["font"]["family"] = "courier"
    pv.rcParams["title"] = "pyansys"
    return


def _check_has_ansys():
    """Safely wraps check_valid_ansys

    Returns
    -------
    has_ansys : bool
        True when this local installation has ANSYS installed in a
        standard location.
    """
    from ansys.mapdl.core.launcher import check_valid_ansys
    try:
        return check_valid_ansys()
    except:
        return False


def supress_logging(func):
    """Decorator to suppress logging for a MAPDL instance"""
    def wrapper(*args, **kwargs):
        mapdl = args[0]
        prior_log_level = mapdl._log.level
        if prior_log_level != 'CRITICAL':
            mapdl._set_log_level('CRITICAL')

        out = func(*args, **kwargs)

        if prior_log_level != 'CRITICAL':
            mapdl._set_log_level(prior_log_level)

        return out

    return wrapper


def run_as_prep7(func):
    """Run a MAPDL method at PREP7 and always revert to the prior processor"""
    def wrapper(*args, **kwargs):
        mapdl = args[0]
        if hasattr(mapdl, '_mapdl'):
            mapdl = mapdl._mapdl
        prior_processor = mapdl.parameters.routine
        if prior_processor != 'PREP7':
            mapdl.prep7()

        out = func(*args, **kwargs)

        if prior_processor == 'Begin level':
            mapdl.finish()
        elif prior_processor != 'PREP7':
            mapdl.run('/%s' % prior_processor)

        return out
    return wrapper


def threaded(fn):
    """ calls a function using a thread """
    def wrapper(*args, **kwargs):
        thread = Thread(target=fn, args=args, kwargs=kwargs)
        thread.start()
        return thread
    return wrapper


def threaded_daemon(fn):
    """Calls a function using a daemon thread."""
    def wrapper(*args, **kwargs):
        thread = Thread(target=fn, args=args, kwargs=kwargs)
        thread.daemon = True
        thread.start()
        return thread
    return wrapper


def chunks(l, n):
    """ Yield successive n-sized chunks from l """
    for i in range(0, len(l), n):
        yield l[i:i + n]


def unique_rows(a):
    """ Returns unique rows of a and indices of those rows """
    if not a.flags.c_contiguous:
        a = np.ascontiguousarray(a)

    b = a.view(np.dtype((np.void, a.dtype.itemsize * a.shape[1])))
    _, idx, idx2 = np.unique(b, True, True)

    return a[idx], idx, idx2


def creation_time(path_to_file):
    """The file creation time.

    Try to get the date that a file was created, falling back to when it was
    last modified if that isn't possible.
    See http://stackoverflow.com/a/39501288/1709587 for explanation.
    """
    if platform.system() == 'Windows':
        return os.path.getctime(path_to_file)
    else:
        stat = os.stat(path_to_file)
        try:
            return stat.st_birthtime
        except AttributeError:
            # We're probably on Linux. No easy way to get creation dates here,
            # so we'll settle for when its content was last modified.
            return stat.st_mtime


def last_created(filenames):
    """Return the last created file given a list of filenames

    If all filenames have the same creation time, then return the last filename.
    """
    ctimes = [creation_time(filename) for filename in filenames]
    idx = np.argmax(ctimes)
    if len(set(ctimes)):
        return filenames[-1]

    return filenames[idx]


def create_temp_dir(tmpdir=None):
    """Create a new unique directory at a given temporary directory"""
    if tmpdir is None:
        tmpdir = tempfile.gettempdir()
    elif not os.path.isdir(tmpdir):
        os.makedirs(tmpdir)

    # running into a rare issue with MAPDL on Windows with "\n" being
    # treated literally.
    letters = string.ascii_lowercase.replace('n', '')
    path = os.path.join(tmpdir, random_string(10, letters))

    # in the *rare* case of a duplicate path
    while os.path.isdir(path):
        path = os.path.join(tempfile.gettempdir(), random_string(10, letters))

    try:
        os.mkdir(path)
    except:
        raise RuntimeError('Unable to create temporary working '
                           'directory %s\n' % path +
                           'Please specify run_location=')

    return path


def no_return(fn):
    """Decorator to return nothing from the wrapped function"""
    def wrapper(*args, **kwargs):
        fn(*args, **kwargs)
    return wrapper



"""Contains the MapdlMath classes, allowing for math operations within
mapdl from Python.  """
import os
import weakref
import string
import random
from enum import Enum

import numpy as np
from ansys.grpc.mapdl import ansys_kernel_pb2 as anskernel
from ansys.grpc.mapdl import mapdl_pb2 as pb_types

from .parameters import interp_star_status
from .errors import ANSYSDataTypeError, protect_grpc
from .mapdl_grpc import MapdlGrpc
from .common_grpc import (parse_chunks, ANSYS_VALUE_TYPE,
                          DEFAULT_CHUNKSIZE, DEFAULT_FILE_CHUNK_SIZE)


MYCTYPE = {np.int32: 'I',
           np.int64: 'L',
           np.single: 'S',
           np.double: 'D',
           np.complex64: 'C',
           np.complex128: 'Z'}


NP_VALUE_TYPE = {value: key for key, value in ANSYS_VALUE_TYPE.items()}

# for windows LONG vs INT32
if os.name == 'nt':
    NP_VALUE_TYPE[np.intc] = 1


def id_generator(size=6, chars=string.ascii_uppercase):
    """Generate a random string"""
    return ''.join(random.choice(chars) for _ in range(size))


class ObjType(Enum):
    """Generic APDLMath Object ( Shared features between Vec Mat and
    Solver components """
    GEN = 1
    VEC = 2
    DMAT = 3
    SMAT = 4


def get_nparray_chunks(name, array, chunk_size=DEFAULT_FILE_CHUNK_SIZE):
    """Serializes a numpy array into chunks"""
    stype = NP_VALUE_TYPE[array.dtype.type]
    arr_sz = array.size
    i = 0  # position counter
    byte_array = array.tobytes()
    while i < len(byte_array):
        piece = byte_array[i: i + chunk_size]
        chunk = anskernel.Chunk(payload=piece, size=len(piece))
        yield pb_types.SetVecDataRequest(vname=name, stype=stype, size=arr_sz,
                                         chunk=chunk)
        i += chunk_size



class MapdlMath():
    """Abstract mapdl math class.  Created from a ``Mapdl`` instance.

    Examples
    --------
    Create an instance from an existing mapdl instance

    >>> import ansys
    >>> mapdl = ansys.Mapdl()
    >>> mm = mapdl.math()

    Alternatively:

    >>> from ansys.mapdl.math import MapdlMath
    >>> mm = MapdlMath(mapdl)  # from mapdl above

    Vector addition

    >>> v1 = mm.ones(10)
    >>> v2 = mm.ones(10)
    >>> v3 = v1 + v2

    Matrix multiplcation

    >>> v1 = mm.ones(10)
    >>> m1 = mm.rand(10, 10)
    >>> v2 = m1*v1

    """

    def __init__(self, mapdl):
        if not isinstance(mapdl, MapdlGrpc):
            raise TypeError('``mapdl`` must be a MapdlGrpc instance')
        self._mapdl_weakref = weakref.ref(mapdl)

    @property
    def _mapdl(self):
        """Return the weakly referenced instance of mapdl"""
        return self._mapdl_weakref()

    def free(self):
        """Delete all vectors"""
        self._mapdl.run("*FREE,ALL")

    def __repr__(self):
        return self._status

    def status(self):
        """Print out the status of all APDLMath Objects"""
        print(self._status)

    @property
    def _status(self):
        """Print out the status of all APDLMath Objects"""
        return self._mapdl.run("*STATUS,MATH")

    def vec(self, size=0, dtype=np.double, init=None, name=None):
        """Create a vector

        Parameters
        ----------
        size : int
            Size of the vector

        dtype : np.dtype, optional
            Datatype of the vector.  Must be either np.int32, np.int64, or np.double

        init : str, optional
            Initialization options.  Can be ones, zeros, or rand.

        Returns
        -------
        vector : ansys.mapdl.math.AnsVec
            ANSYS Vector
        """
        if dtype not in MYCTYPE:
            raise ANSYSDataTypeError

        if not name:
            name = id_generator()
            self._mapdl.run("*VEC," + name + "," + MYCTYPE[dtype] + ",ALLOC," + str(size))
            return AnsVec(name, self._mapdl, dtype, init)
        else:
            return AnsVec(name, self._mapdl)
    
    def mat(self, nrow=0, ncol=0, dtype=np.double, init=None, name=None):
        """Create matrix

        Parameters
        ----------
        nrow : int
            Number of rows.

        ncol : int
            Number of columns.

        dtype : np.dtype, optional
            Datatype of the vector.  Must be either np.int32, np.int64, or np.double

        init : str, optional
            Initialization options.  Can be ones, zeros, or rand.

        Returns
        -------
        matrix : ansys.math.AnsMat
            ANSYS matrix
        """
        if dtype not in MYCTYPE:
            raise ValueError('Invalid datatype.  Must be one of the following:\n' +
                             'np.int32, np.int64, or np.double')

        if not name:
            name = id_generator()
            self._mapdl.run(f'*DMAT,{name},{MYCTYPE[dtype]},ALLOC,{nrow},{ncol}')
            mat = AnsDenseMat(name, self._mapdl)

            if (init == "rand"):
                mat.rand()
            elif (init == "ones"):
                mat.ones()
            elif (init == "zeros"):
                mat.zeros()
            elif init is not None:
                raise ValueError("Invalid init method")
        else:
            info = self._mapdl._data_info(name)
            mtype = info.objtype
            if mtype == 2:
                return AnsDenseMat(name, self._mapdl)
            else:
                return AnsSparseMat(name, self._mapdl)

        return mat

    def zeros(self, nrow, ncol=None, dtype=np.double):
        """Create a vector or matrix containing all zeros

        nrow : int
            Number of rows.

        ncol : int, optional
            Number of columns.  If specified, returns a matrix.

        dtype : np.dtype, optional
            Datatype of the vector.  Must be either np.int32, np.int64, or np.double

        Returns
        -------
        ans_vec_mat : ansys.AnsVec or ansys.AnsMat
            ANSYS vector or matrix depending on if "ncol" is specified

        Examples
        --------
        Create a zero vector

        >>> import ansys
        >>> mapdl = ansys.Mapdl()
        >>> mm = mapdl.math()
        >>> vec = mm.zeros(10)

        Create a zero matrix

        >>> mat = mm.zeros(10, 10)
        """
        if dtype not in MYCTYPE:
            raise ANSYSDataTypeError

        if not ncol:
            return self.vec(nrow, dtype, init="zeros")
        else:
            return self.mat(nrow, ncol, dtype, init="zeros")

    def ones(self, nrow, ncol=None, dtype=np.double):
        """Create a vector or matrix containing all ones

        nrow : int
            Number of rows.

        ncol : int, optional
            Number of columns.  If specified, returns a matrix.

        dtype : np.dtype, optional
            Datatype of the vector.  Must be either np.int32, np.int64, or np.double

        Returns
        -------
        ans_vec_mat : ansys.AnsVec or ansys.AnsMat
            ANSYS vector or matrix depending on if "ncol" is specified

        Examples
        --------
        Create a ones vector

        >>> import ansys
        >>> mapdl = ansys.Mapdl()
        >>> mm = mapdl.math()
        >>> vec = mm.ones(10)

        Create a ones matrix

        >>> mat = mm.ones(10, 10)
        """

        if not ncol:
            return self.vec(nrow, dtype, init="ones")
        else:
            return self.mat(nrow, ncol, dtype, init="ones")

    def rand(self, nrow, ncol=None, dtype=np.double):
        """Create a vector or matrix containing all random values

        nrow : int
            Number of rows.

        ncol : int, optional
            Number of columns.  If specified, returns a matrix.

        dtype : np.dtype, optional
            Datatype of the vector.  Must be either np.int32, np.int64, or np.double

        Returns
        -------
        ans_vec_mat : ansys.AnsVec or ansys.AnsMat
            ANSYS vector or matrix depending on if "ncol" is specified

        Examples
        --------
        Create a random vector

        >>> import ansys
        >>> mapdl = ansys.Mapdl()
        >>> mm = mapdl.math()
        >>> vec = mm.rand(10)

        Create a random matrix

        >>> mat = mm.rand(10, 10)
        """
        if not ncol:
            return self.vec(nrow, dtype, init="rand")
        else:
            return self.mat(nrow, ncol, dtype, init="rand")

    def matrix(self, matrix, name=None, triu=False):
        """Send a scipy matrix or numpy array to MAPDL.

        Parameters
        ----------
        matrix : np.ndarray
            Numpy array to send as a matrix to MAPDL.

        name : str, optional
            APDLMath matrix name.  Optional and defaults to a random name.

        triu: bool, optional
            ``True`` when the matrix is upper triangular, ``False``
            when unsymmetric.

        Returns
        -------
        ans_mat : MapdlMath.AnsMat
            MapdlMath matrix.

        Examples
        --------
        Generate a random sparse matrix

        >>> from scipy import sparse
        >>> sz = 5000
        >>> mat = sparse.random(sz, sz, density=0.05, format='csr')
        >>> ans_mat = mm.matrix(mat, name)
        >>> ans_mat
        APDLMath Matrix 5000 x 5000

        Read the matrix back to Python

        >>> ans_mat.asarray()
        <500x5000 sparse matrix of type '<class 'numpy.float64'>'
                with 1250000 stored elements in Compressed Sparse Row format>

        """
        if name is None:
            name = id_generator()
        elif not isinstance(name, str):
            raise TypeError('``name`` parameter must be a string')

        self._set_mat(name, matrix, triu)
        return AnsSparseMat(name, self._mapdl)

    def load_matrix_from_file(self, type=np.double, fname="file.full", matId="STIFF"):
        """Load a matrix from a file"""
        name = id_generator()
        self._mapdl._log.info(f"Calling MAPDL to extract the %s matrix from %s",
                              matId, fname)
        self._mapdl.run(f"*SMAT,{name},{MYCTYPE[type]},IMPORT,FULL,{fname},{matId}")
        return AnsSparseMat(name, self._mapdl)

    def stiff(self, dtype=np.double, fname="file.full"):
        """Load the stiffness matrix from the full file

        Examples
        --------
        >>> k = mapdl.math.stiff()
        APDLMATH Matrix 60 x 60

        Convert to a scipy array

        >>> mat = k.asarray()
        >>> mat
        <60x60 sparse matrix of type '<class 'numpy.float64'>'
            with 1734 stored elements in Compressed Sparse Row format>
        """
        return self.load_matrix_from_file(dtype, fname, "STIFF")

    def mass(self, dtype=np.double, fname="file.full"):
        """Load the mass matrix from the full file

        Examples
        --------
        >>> m = mapdl.math.mass()
        >>> m
        APDLMATH Matrix 60 x 60

        Convert to a scipy array

        >>> mat = m.asarray()
        >>> mat
        <60x60 sparse matrix of type '<class 'numpy.float64'>'
            with 1734 stored elements in Compressed Sparse Row format>
        """
        return self.load_matrix_from_file(dtype, fname, "MASS")

    def damp(self, dtype=np.double, fname="file.full"):
        """Load the damping matrix from the full file"""
        return self.load_matrix_from_file(dtype, fname, "DAMP")

    def getVec(self, dtype=np.double, fname="file.full", matId="RHS"):
        """Load a vector from a file"""
        name = id_generator()
        self._mapdl._log.info("Call MAPDL to extract the %s vector from the file %s",
                              matId, fname)
        self._mapdl.run(f"*VEC,{name},{MYCTYPE[dtype]},IMPORT,FULL,{fname},{matId}")
        return AnsVec(name, self._mapdl)

    def set_vec(self, vname, data):
        """Push a numpy array or Python list to the MAPDL Memory
        Workspace.

        Parameters
        ----------
        vname : str
            APDLMath vector name

        data : np.ndarray, list
            Numpy array or Python list to push to MAPDL.  Must be 1
            dimensional.

        Returns
        -------
        vec : ansys.mapdl.math.AnsVec
            MAPDL vector instance generated from the pushed vector.

        Examples
        --------
        >>> data = np.random.random(10)
        >>> vec = mm.set_vec('NewVec', data)
        >>> np.isclose(vec.asarray(), data)
        True
        """
        self._set_vec(vname, data)
        return AnsVec(vname, self._mapdl)

    def rhs(self, type=np.double, fname="file.full"):
        return self.getVec(type, fname, "RHS")

    def eigs(self, nev, k, m=None, c=None, phi=None, algo=None,
             fmin=None, fmax=None):
        """Solve an eigenproblem

        Parameters
        ----------
        nev : int
            Number of eigenvalues to compute.

        K : ansys.AnsMat
            An array representing the operation ``A * x`` where A is a
            square matrix.

        M : ansys.AnsMat, optional
            An array representing the operation ``M * x`` for the
            generalized eigenvalue problem:

            ``K * x = M * x``

        Examples
        --------
        Solve an eigenproblem using the mass and stiffness matrices
        stored from a prior ansys run.

        >>> k = mm.stiff()
        >>> m = mm.mass()
        >>> nev = 10
        >>> a = mm.mat(k.nrow, nev)
        >>> ev = mm.eigs(nev, k, m, phi=a)
        """
        if not fmin:
            fmin = ''
        if not fmax:
            fmax = ''

        cid = ''

        if not c:
            if (k.sym() and m.sym()):
                if not algo:
                    algo = "LANB"
            else:
                algo = "UNSYM"
        else:
            cid = c.id
            algo = "DAMP"

        self._mapdl.run("/SOLU")
        self._mapdl.run("antype,modal")
        self._mapdl.run("modopt," + algo + "," + str(nev) + "," + str(fmin) + "," + str(fmax))
        ev = self.vec()

        phistr = '' if not phi else phi.id
        self._mapdl.run("*EIG," + k.id + "," + m.id + "," + cid + "," + ev.id + "," + phistr)

        return ev

    def dot(self, vec_a, vec_b):
        """Dot product between two ANSYS vector objects.

        Parameters
        ----------
        vec_a : ansys.mapdl.math.AnsVec
            Ansys vector object.

        vec_b : ansys.mapdl.math.AnsVec
            Ansys vector object.

        Returns
        -------
        dot_product : float
            Dot product between the two vectors.

        Examples
        --------
        >>> v = mm.ones(10)
        >>> w = mm.ones(10)
        >>> v.dot(w)
        """
        return dot(vec_a, vec_b)

    def add(self, obj1, obj2):
        """Add two APDLMath vectors or matrices.

        Parameters
        ----------
        obj1 : ansys.mapdl.math.AnsVec or ansys.mapdl.math.AnsMat
            Ansys object.

        obj2 : ansys.mapdl.math.AnsVec or ansys.mapdl.math.AnsMat
            Ansys object.

        Returns
        -------
        objout : ansys.mapdl.math.AnsVec or ansys.mapdl.math.AnsMat
            Sum of the two input objects.  Type of the output matches type of the input.
            Sum of the two vectors/matrices.

        Examples
        --------
        >>> v = mm.ones(10)
        >>> w = mm.ones(10)
        >>> x = mm.add( v, w)
        """
        return obj1 + obj2

    def subtract(self, obj1, obj2):
        """Subtract two ANSYS vectors or matrices.

        Parameters
        ----------
        obj1 : ansys.mapdl.math.AnsVec or ansys.mapdl.math.AnsMat
            Ansys object.

        obj2 : ansys.mapdl.math.AnsVec or ansys.mapdl.math.AnsMat
            Ansys object.

        Returns
        -------
        objout : ansys.mapdl.math.AnsVec or ansys.mapdl.math.AnsMat
            Difference of the two input vectors or matrices.  Type of the output
            matches the type of the input.

        Examples
        --------
        >>> v = mm.ones(10)
        >>> w = mm.ones(10)
        >>> x = mm.add( v, w)
        """
        return obj1 - obj2

    
    def factorize(self, mat):
        """Factorize a matrix

        Parameters
        ----------
        mat : ansys.mapdl.math.AnsMat
            A ansys.mapdl.math matrix

        Examples
        --------
        >>> mm = mapdl.math
        >>> dim = 1000
        >>> m2 = mm.rand(dim, dim)
        >>> m3 = m2.copy()
        >>> mat = mm.factorize(m2)
        
        """
        solver = AnsSolver(id_generator(), self._mapdl)
        solver.factorize(mat)
        return solver

    def norm(self, obj, ord="nrm2"):
        """ Matrix or vector norm

        Parameters
        ----------
        obj : ansys.mapdl.math.AnsMat or ansys.mapdl.math.AnsVec
        ord : Order of the norm. nrm2(default), nrminf, nrm1

        Examples
        --------
        >>> mm = mapdl.math
        >>> dim = 1000
        >>> m2 = mm.rand(dim, dim)
        >>> nrm = mm.norm( m2)
        """
        obj.norm()

    @protect_grpc
    def _set_vec(self, vname, arr, dtype=None, chunk_size=DEFAULT_CHUNKSIZE):
        """Transfer a numpy array to MAPDL as a MAPDL Math vector.

        Parameters
        ----------
        vname : str
            Vector parameter name.  Character ":" is not allowed.

        arr : np.ndarray
            Numpy array to upload

        dtype : np.dtype, optional
            Type to upload array as.  Defaults to the current array type.

        chunk_size : int, optional
            Chunk size in bytes.  Must be less than 4MB.

        """
        if ':' in vname:
            raise ValueError('The character ":" is not permitted in a MAPDL MATH'
                             ' vector parameter name')
        if not isinstance(arr, np.ndarray):
            arr = np.asarray(arr)

        if dtype is not None:
            if arr.dtype != dtype:
                arr = arr.astype(dtype)

        if arr.dtype not in list(NP_VALUE_TYPE.keys()):
            dtypes = list(NP_VALUE_TYPE.keys())
            if None in dtypes:
                dtypes.remove(None)
            raise TypeError('Invalid array datatype.  Must be one of the following:\n'
                            + ('\n'.join([str(dtype) for dtype in dtypes])))
        chunks_generator = get_nparray_chunks(vname, arr, chunk_size)
        self._mapdl._stub.SetVecData(chunks_generator)

    @protect_grpc
    def _set_mat(self, mname, arr, sym, dtype=None, chunk_size=DEFAULT_CHUNKSIZE):
        """Transfer a scipy array to MAPDL as a MAPDL Math matrix

        Parameters
        ----------
        mname : str
            Matrix parameter name.  Character ":" is not allowed.

        arr : 
            SciPy array to upload

        sym : bool
            ``True`` when sparse matrix is symmetric.

        dtype : np.dtype, optional
            Type to upload array as.  Defaults to the current array type.

        chunk_size : int, optional
            Chunk size in bytes.  Must be less than 4MB.

        """
        if ':' in mname:
            raise ValueError('The character ":" is not permitted in a MAPDL MATH'
                             ' matrix parameter name')
        if not len(mname):
            raise ValueError('Empty MAPDL matrix name not permitted')

        if isinstance(arr, np.ndarray):
            if arr.ndim == 1:
                raise ValueError('Input appears to be an array.  '
                                 'Use ``set_vec`` instead.)')
            if arr.ndim > 2:
                raise ValueError('Dense arrays must be 2 dimensional')

            from scipy import sparse
            arr = sparse.csr_matrix(arr)

        if dtype is not None:
            if arr.data.dtype != dtype:
                arr.data = arr.data.astype(dtype)

        if arr.data.dtype not in list(NP_VALUE_TYPE.keys()):
            dtypes = list(NP_VALUE_TYPE.keys())
            if None in dtypes:
                dtypes.remove(None)
            raise TypeError('Invalid array datatype.  Must be one of the following:\n'
                            + ('\n'.join([str(dtype) for dtype in dtypes])))

        if arr.shape[0] != arr.shape[1]:
            raise ValueError('APDLMath only supports square matrices')

        # The data vector
        dataname = f'{mname}_DATA'
        self.set_vec(dataname, arr.data)

        # indptr vector
        indptrname = f'{mname}_IND'
        indv = arr.indptr.astype('int64') + 1
        self.set_vec(indptrname, indv)

        # indices vector
        indxname = f'{mname}_PTR'
        idx = arr.indices + 1
        self.set_vec(indxname, idx)

        flagsym = 'FALSE'
        if sym is True:
            flagsym = 'TRUE'

        self._mapdl.run(f'*SMAT,{mname},D,ALLOC,CSR,{indptrname},{indxname},{dataname},{flagsym}')


class ApdlMathObj:
    def __init__(self, id, mapdl, type=ObjType.GEN):
        self.id = id
        self._mapdl = mapdl
        self.type = type

    #def __del__(self):
    #    self._mapdl._log.debug("Deleting the MAPDL Vector Object")
    #    self._mapdl.run("*FREE," + self.id)

    def __repr__(self):
        return "APDLMath Object %s" % str(self.id)

    def __str__(self):
        return self._mapdl.run("*PRINT," + self.id)

    def copy(self):
        """Returns the name of the copy of this object"""
        name = id_generator()  # internal name of the new vector
        if self.type == ObjType.VEC:
            acmd = "*VEC"
        elif self.type == ObjType.DMAT:
            acmd = "*DMAT"
        elif self.type == ObjType.SMAT:
            acmd = "*SMAT"
        else:
            raise TypeError("Unknown obj type: operation aborted")

        # APDLMath cmd to COPY vin to vout
        self._mapdl.run(acmd + "," + name + ",D,COPY," + self.id)
        return name

    def _init(self, method):
        cmd = "*INIT," + self.id + "," + method
        self._mapdl.run(cmd)

    def zeros(self):
        """Set all values of the vector to zero"""
        return self._init("ZERO")

    def ones(self):
        """Set all values of the vector to one"""
        return self._init("CONST,1")

    def rand(self):
        """Set all values of the vector to a random number"""
        return self._init("RAND")

    def const(self, value):
        """Set all values of the vector to a constant"""
        return self._init("CONST," + str(value))

    def norm(self, nrmtype="nrm2"):
        """Matrix or vector norm.

        Parameters
        ----------
        nrmtype : str, optional
            Mathematical norm to use.  One of:

            - ``'NRM2'``: L2 (Euclidean or SRSS) norm (default).
            - ``'NRM1'``: L1 (absolute sum) norm (vectors only).
            - ``'NRMINF'`` Maximum norm.

        Returns
        -------
        nrm : float
            Norm of the matrix or vector(s).

        Examples
        --------
        >>> mm = mapdl.math
        >>> dim = 1000
        >>> m2 = mm.rand(dim, dim)
        >>> nrm = mm.norm( m2)
        """
        val_name = 'py_val'
        self._mapdl.run("*NRM," + self.id + "," + nrmtype + ",%s" % val_name)
        return self._mapdl.scalar_param(val_name)

    def axpy(self, op, val1, val2):
        """Perform the matrix operation: ``M2= v*M1 + w*M2``"""
        cmd = "*AXPY," + str(val1) + ",0," + op.id + "," + str(val2) + ",0," + self.id
        self._mapdl._log.info(f">> Call Mapdl to perform AXPY operation")
        self._mapdl.run(cmd)
        return self

    def __add__(self, op2):
        opout = self.copy()
        self._mapdl._log.info(">> Call Mapdl to perform AXPY operation")
        self._mapdl.run(f"*AXPY,1,0,{op2.id},1,0,{opout.id}")
        return opout

    def __sub__(self, op2):
        opout = self.copy()
        self._mapdl._log.info("Call Mapdl to perform AXPY operation")
        self._mapdl.run(f"*AXPY,-1,0,{op2.id},1,0,{opout.id}")
        return opout

    def __iadd__(self, op):
        return self.axpy(op, 1, 1)

    def __isub__(self, op):
        return self.axpy(op, -1, 1)

    def __imul__(self, val):
        self._mapdl._log.info("Call Mapdl to scale the object")
        self._mapdl.run(f"*SCAL,{self.id},{val}")
        return self

    def __itruediv__(self, val):
        if val == 0:
            raise ZeroDivisionError('division by zero')
        self._mapdl._log.info("Call Mapdl to 1/scale the object")
        self._mapdl.run(f"*SCAL,{self.id},{1/val}")
        return self

    @property
    @protect_grpc
    def _data_info(self):
        """Return the data type of a parameter"""
        request = pb_types.ParameterRequest(name=self.id)
        return self._stub.GetDataInfo(request)


class AnsVec(ApdlMathObj):
    """APDLMath Vector Object"""
    def __init__(self, id, mapdl, dtype=np.double, init=None):
        ApdlMathObj.__init__(self, id, mapdl, ObjType.VEC)

        if init not in ['ones', 'zeros', 'rand', None]:
            raise ValueError('Invalid init option.  Should be "ones", "zeros", "rand", or None')

        if (init == "rand"):
            self.rand()
        elif (init == "ones"):
            self.ones()
        elif (init == "zeros"):
            self.zeros()

    @property
    def size(self):
        return int(self._mapdl.scalar_param(self.id + "_DIM"))

    def __repr__(self):
        return "APDLMath Vector Size %s" % self.size

    def __getitem__(self, num):
        if num < 0:
            raise ValueError('Negative indices not permitted')
        cmd = "pyval=" + self.id + "(" + str(num+1) + ")"
        self._mapdl.run(cmd)
        return self._mapdl.scalar_param("pyval")

    def __mul__(self, vec2):
        raise AttributeError('Array multiplication is not yet available.  '
                             'For dot product, please use `dot()`')

    def copy(self):
        """Return a copy of this vector"""
        return AnsVec(ApdlMathObj.copy(self), self._mapdl)

    def dot(self, vec):
        """Dot product with another ansys vector object

        Parameters
        ----------
        vec : ansys.mapdl.math.AnsVec
            Ansys vector object.

        Returns
        -------
        dot_product : float
            Dot product between this vector and the other vector.
        """
        if not isinstance(vec, AnsVec):
            raise TypeError('Must be an Ansys vector object')

        cmd = "*DOT," + self.id + "," + vec.id + ",py_val"
        self._mapdl.run(cmd)
        return self._mapdl.scalar_param("py_val")

    def asarray(self):
        """Returns vector as a numpy array

        Examples
        --------
        >>> from ansys.mapdl import Mapdl
        >>> mm = mapdl.math()
        >>> v = mm.ones(10)
        >>> v.asarray()
        [1. 1. 1. 1. 1. 1. 1. 1. 1. 1.]
        """
        return self._mapdl._vec_data(self.id)

    def __array__(self):
        """Allow numpy to access this object as if it was an array"""
        return self.asarray()


class AnsMat(ApdlMathObj):
    """APDLMath Matrix Object"""

    def __init__(self, id, mapdl, type=ObjType.DMAT):
        ApdlMathObj.__init__(self, id, mapdl, type)

    @property
    def nrow(self):
        return int(self._mapdl.scalar_param(self.id + "_ROWDIM"))

    @property
    def ncol(self):
        return int(self._mapdl.scalar_param(self.id + "_COLDIM"))

    @property
    def size(self):
        return self.nrow*self.ncol

    @property
    def shape(self):
        """Returns a numpy-like shape.

        Tuple of (rows and columns)
        """
        return (self.nrow, self.ncol)

    def sym(self):  # BUG this is not always true
        return True

    def asarray(self):
        """Returns vector as a numpy array

        Examples
        --------
        >>> from ansys.mapdl import Mapdl
        >>> mm = mapdl.math()
        >>> v = mm.ones(10)
        >>> v.asarray()
        [1. 1. 1. 1. 1. 1. 1. 1. 1. 1.]
        """
        return self._mapdl._mat_data(self.id)

    def __repr__(self):
        return "APDLMath Matrix %d x %d" % (self.nrow, self.ncol)

    def __mul__(self, vec):
        raise AttributeError('Array multiplication is not yet available.  '
                             'For dot product, please use `dot()`')

    def dot(self, obj):
        """Perform the matrix multiplication by another vector or matrix

        Parameters
        ----------
        obj : ansys.mapdl.math.AnsVec or ansys.mapdl.math.AnsMat
            Ansys object.

        Returns
        -------
        dot_product : ansys.mapdl.math.AnsVec or ansys.mapdl.math.AnsMat
            Matrix multiplication result.

        Examples
        --------
        Multiplication of a matrix and vector

        >>> m1 = mm.rand(10, 10)
        >>> v1 = mm.rand(10)
        >>> v2 = m1.dot(v1)
        >>> assert np.allclose(m1.asarray() @ v1.asarray(), v2)

        """
        name = id_generator()  # internal name of the new vector/matrix
        info = self._mapdl._data_info(self.id)
        dtype = ANSYS_VALUE_TYPE[info.stype]
        if obj.type == ObjType.VEC:
            self._mapdl.run(f"*VEC,{name},{MYCTYPE[dtype]},ALLOC,{info.size1}")
            objout = AnsVec(name, self._mapdl)
        else:
            self._mapdl.run(f"*DMAT,{name},{MYCTYPE[dtype]},ALLOC,{info.size1},{info.size2}")
            objout = AnsDenseMat(name, self._mapdl)
            
        cmd = f"*MULT,{self.id},,{obj.id},,{name}"
        self._mapdl._log.info("Call Mapdl to perform MV Product")
        self._mapdl.run(cmd)
        return objout

    def __getitem__(self, num):
        name = id_generator()
        cmd = "*VEC," + name + ",D,LINK," + self.id + "," + str(num+1)
        self._mapdl.run(cmd)
        return AnsVec(name, self._mapdl)

    @property
    def T(self):
        """Returns the transpose of a MAPDL matrix.

        Examples
        --------
        >>> from ansys.mapdl import Mapdl
        >>> mm = mapdl.math()
        >>> mat = mm.rand(2, 3)
        >>> mat_t = mat.T

        """
        info = self._mapdl._data_info(self.id)

        if info.objtype == 2:
            objtype = "*DMAT"
        else:
            objtype = "*SMAT"

        dtype = ANSYS_VALUE_TYPE[info.stype]
        name = id_generator()
        self._mapdl._log.info("Call MAPDL to transpose")
        self._mapdl.run(f"{objtype},{name},{MYCTYPE[dtype]},COPY,{self.id},TRANS")
        if info.objtype == 2:
            mat = AnsDenseMat(name, self._mapdl)
        else:
            mat = AnsSparseMat(name, self._mapdl)
        return mat


class AnsDenseMat(AnsMat):
    """Dense APDLMath Matrix"""

    def __init__(self, uid, mapdl):
        AnsMat.__init__(self, uid, mapdl, ObjType.DMAT)

    def __array__(self):
        """Allow numpy to access this object as if it was an array"""
        return self.asarray()

    def __repr__(self):
        return f'Dense APDLMath Matrix ({self.nrow}, {self.ncol})'

    def copy(self):
        """Return a copy of this matrix"""
        return AnsDenseMat(ApdlMathObj.copy(self), self._mapdl)


class AnsSparseMat(AnsMat):
    """Sparse APDLMath Matrix"""

    def __init__(self, uid, mapdl):
        AnsMat.__init__(self, uid, mapdl, ObjType.SMAT)

    def __repr__(self):
        return f'Sparse APDLMath Matrix ({self.nrow}, {self.ncol})'

    def copy(self):
        """Return a copy of this matrix"""
        return AnsSparseMat(ApdlMathObj.copy(self), self._mapdl)

    def todense(self):
        """Return this array as a dense np.ndarray"""
        return self.asarray().todense()

    def __array__(self):
        """Allow numpy to access this object as if it was an array"""
        return self.todense()


class AnsSolver(ApdlMathObj):
    """APDLMath Solver Class"""
    def __repr__(self):
        return "APDLMath Linear Solver"

    # def __del__(self):
    #     self._mapdl._log.debug("Deleting the MAPDL Solver Object")
    #     self._mapdl.run(f"*FREE,{self.id}")

    def factorize(self, M, Algo=None):
        """Factorize a matrix

        Performs the numerical factorization of a linear solver system.

        Parameters
        ----------
        """
        if not Algo:
            if (M.type == ObjType.DMAT):
                Algo = "LAPACK"
            elif (M.type == ObjType.SMAT):
                Algo = "DSP"
        self._mapdl.run(f"*LSENGINE,{Algo},{self.id},{M.id}")
        self._mapdl._log.info("Factorizing using the %s package", Algo)
        self._mapdl.run("*LSFACTOR," + self.id)

    def solve(self, B, X=None):
        if not X:
            X = B.copy()
        self._mapdl._log.info("Solving")
        self._mapdl.run("*LSBAC," + self.id + "," + B.id + "," + X.id)
        return X


def rand(obj):
    """Set all values of a mapdl math object to random values

    Parameters
    ----------
    obj : ansys.math.MapdlMath object
        MapdlMath object
    """
    obj._mapdl.cmd("*INIT," + obj.id + ",RAND")


def solve(Mat, B, X=None, Algo=None):
    Solver = AnsSolver(id_generator(), Mat._mapdl)
    Solver.factorize(Mat, Algo)
    if not X:
        X = B.copy()
    X = Solver.solve(B, X)

    del Solver
    return X


def dot(vec1, vec2):
    """Dot product between two ANSYS vector objects

    Parameters
    vec1 : ansys.mapdl.math.AnsVec
        Ansys vector object.

    vec1 : ansys.mapdl.math.AnsVec
        Ansys vector object.

    Returns
    -------
    dot_product : float
        Dot product between the two vectors

    """
    if (vec1.type != ObjType.VEC or vec2.type != ObjType.VEC):
        raise Exception("Both objects must be ANSYS vectors")

    mapdl = vec1._mapdl
    mapdl.run("*DOT," + vec1.id + "," + vec2.id + ",py_val")
    return mapdl.scalar_param("py_val")


def interp_star_status(status, ignore_hidden=True):
    """Interprets STATUS command output from MAPDL

    Parameters
    ----------
    status : str
        Output from APDL *STATUS

    Returns
    -------
    parameters : dict
        Dictionary of parameters.
    """
    parameters = {}
    header = 'Name                   Type            '
    for line in status.splitlines():
        if header in line:
            continue

        try:
            name, mtype, size_mbytes, dim, workspace = line.split()
        except:
            continue

        if ignore_hidden:
            if name[-5:] == '_DATA':
                continue
            elif name[-4:] == '_IND':
                continue
            elif name[-4:] == '_PTR':
                continue

        parameters[name] = {'type': mtype,
                            'size mb': size_mbytes,
                            'dim': dim,
                            'workspace': workspace}

    return parameters



"""Plotting helper for MAPDL using pyvista"""
import pyvista as pv
import numpy as np

from ansys.mapdl.core.misc import unique_rows


def general_plotter(title, meshes, points, labels,
                    cpos=None,
                    show_bounds=False, show_axes=True,
                    background=None, off_screen=None,
                    screenshot=False,
                    window_size=None,
                    notebook=None,
                    # add_mesh kwargs:
                    color='w',
                    show_edges=None, edge_color=None, point_size=5.0,
                    line_width=None, opacity=1.0, flip_scalars=False,
                    lighting=None, n_colors=256,
                    interpolate_before_map=True, cmap=None,
                    render_points_as_spheres=False, render_lines_as_tubes=False,
                    stitle=None,
                    smooth_shading=None,
                    # labels kwargs
                    font_size=None,
                    font_family=None,
                    text_color=None):
    """General pyansys plotter for APDL geometry and meshes.

    Parameters
    ----------
    cpos : list(tuple(floats)), str
        The camera position to use.  You can either use a saved camera
        position or specify one of the following strings:

        - ``"xy"``
        - ``"xz"``
        - ``"yz"``
        - ``"yx"``
        - ``"zx"``
        - ``"zy"``
        - ``"iso"``

    off_screen : bool, optional
        Renders off screen when ``True``.  Useful for automated
        screenshots.

    window_size : list, optional
        Window size in pixels.  Defaults to ``[1024, 768]``

    notebook : bool, optional
        When True, the resulting plot is placed inline a jupyter
        notebook.  Assumes a jupyter console is active.  Automatically
        enables off_screen.

    show_bounds : bool, optional
        Shows mesh bounds when ``True``.

    show_axes : bool, optional
        Shows a vtk axes widget.  Enabled by default.

    screenshot : str or bool, optional
        Saves screenshot to file when enabled.

    color : string or 3 item list, optional, defaults to white
        Use to make the entire mesh have a single solid color.  Either
        a string, RGB list, or hex color string.  For example:
        ``color='white'``, ``color='w'``, ``color=[1, 1, 1]``, or
        ``color='#FFFFFF'``. Color will be overridden if scalars are
        specified.

    show_edges : bool, optional
        Shows the edges of a mesh.  Does not apply to a wireframe
        representation.

    edge_color : string or 3 item list, optional, defaults to black
        The solid color to give the edges when ``show_edges=True``.
        Either a string, RGB list, or hex color string.

    point_size : float, optional
        Point size of any nodes in the dataset plotted. Also applicable
        when style='points'. Default ``5.0``

    line_width : float, optional
        Thickness of lines.  Only valid for wireframe and surface
        representations.  Default None.

    opacity : float, str, array-like
        Opacity of the mesh. If a single float value is given, it will be
        the global opacity of the mesh and uniformly applied everywhere -
        should be between 0 and 1. A string can also be specified to map
        the scalars range to a predefined opacity transfer function
        (options include: 'linear', 'linear_r', 'geom', 'geom_r').
        A string could also be used to map a scalars array from the mesh to
        the opacity (must have same number of elements as the
        ``scalars`` argument). Or you can pass a custom made transfer
        function that is an array either ``n_colors`` in length or shorter.

    n_colors : int, optional
        Number of colors to use when displaying scalars. Defaults to 256.
        The scalar bar will also have this many colors.

    cmap : str, list, optional
       Name of the Matplotlib colormap to us when mapping the ``scalars``.
       See available Matplotlib colormaps.  Only applicable for when
       displaying ``scalars``. Requires Matplotlib to be installed.
       ``colormap`` is also an accepted alias for this. If ``colorcet`` or
       ``cmocean`` are installed, their colormaps can be specified by name.

        You can also specify a list of colors to override an
        existing colormap with a custom one.  For example, to
        create a three color colormap you might specify
        ``['green', 'red', 'blue']``

    render_points_as_spheres : bool, optional
        Render points as spheres.

    render_lines_as_tubes : bool, optional
        Renders lines as tubes.

    smooth_shading : bool, optional
        Smoothly render curved surfaces when plotting.  Not helpful
        for all meshes.
    """
    if notebook:
        off_screen = True

    pl = pv.Plotter(off_screen=off_screen, notebook=notebook)

    if background:
        pl.set_background(background)

    for point in points:
        pl.add_points(point['points'],
                      scalars=point.get('scalars', None),
                      color=color,
                      show_edges=show_edges, edge_color=edge_color,
                      point_size=point_size, line_width=line_width,
                      opacity=opacity, flip_scalars=flip_scalars,
                      lighting=lighting, n_colors=n_colors,
                      interpolate_before_map=interpolate_before_map,
                      cmap=cmap, render_points_as_spheres=render_points_as_spheres,
                      render_lines_as_tubes=render_lines_as_tubes)

    for mesh in meshes:
        pl.add_mesh(mesh['mesh'],
                    scalars=mesh.get('scalars', None),
                    stitle=mesh.get('stitle', None),
                    color=mesh.get('color', color),
                    show_edges=show_edges, edge_color=edge_color,
                    smooth_shading=smooth_shading,
                    point_size=point_size, line_width=line_width,
                    opacity=opacity, flip_scalars=flip_scalars,
                    lighting=lighting, n_colors=n_colors,
                    interpolate_before_map=interpolate_before_map,
                    cmap=cmap, render_points_as_spheres=render_points_as_spheres,
                    render_lines_as_tubes=render_lines_as_tubes)

    for label in labels:
        # verify points are not duplicates
        points, idx, _ = unique_rows(np.array(label['points']))
        labels = np.array(label['labels'])[idx].tolist()

        pl.add_point_labels(points,
                            labels,
                            show_points=False, shadow=False,
                            font_size=font_size,
                            font_family=font_family,
                            text_color=text_color)

    if stitle is not None:
        pl.add_scalar_bar(title=stitle)

    if cpos:
        pl.camera_position = cpos

    if show_bounds:
        pl.show_bounds()

    if show_axes:
        pl.show_axes()

    if screenshot:
        pl.show(title=title, auto_close=False, window_size=window_size)
        pl.screenshot(screenshot)
    else:
        pl.show()

    return pl.camera_position



"""Functions to download sample datasets from the pymapdl data repository
"""
import shutil
import os
import urllib.request
import zipfile

import ansys.mapdl.core as pymapdl


def get_ext(filename):
    """Extract the extension of the filename"""
    ext = os.path.splitext(filename)[1].lower()
    return ext


def delete_downloads():
    """Delete all downloaded examples to free space or update the files"""
    shutil.rmtree(pymapdl.EXAMPLES_PATH)
    os.makedirs(pymapdl.EXAMPLES_PATH)
    return True


def _decompress(filename):
    zip_ref = zipfile.ZipFile(filename, 'r')
    zip_ref.extractall(pymapdl.EXAMPLES_PATH)
    return zip_ref.close()


def _get_file_url(filename):
    return f'https://github.com/akaszynski/pyansys-data/raw/master/data/{filename}'


def _retrieve_file(url, filename):
    # First check if file has already been downloaded
    local_path = os.path.join(pymapdl.EXAMPLES_PATH, os.path.basename(filename))
    local_path_no_zip = local_path.replace('.zip', '')
    if os.path.isfile(local_path_no_zip) or os.path.isdir(local_path_no_zip):
        return local_path_no_zip, None

    # grab the correct url retriever
    urlretrieve = urllib.request.urlretrieve

    # Perform download
    saved_file, resp = urlretrieve(url)
    shutil.move(saved_file, local_path)
    if get_ext(local_path) in ['.zip']:
        _decompress(local_path)
        local_path = local_path[:-4]
    return local_path, resp


def download_file(filename):
    return _download_file(filename)


def _download_file(filename):
    url = _get_file_url(filename)
    return _retrieve_file(url, filename)


def _download_and_read(filename):
    saved_file, _ = _download_file(filename)
    if saved_file[-3:] == 'cdb':
        return pymapdl.Archive(saved_file)
    else:
        return pymapdl.read_binary(saved_file)


###############################################################################
# def download_verification_result(index):
#     """Download a verification manual result file"""
#     return _download_and_read('vm%d.rst' % index)



import tempfile
import weakref

import os
import numpy as np

from ansys.mapdl.reader._reader import write_array

import ansys.mapdl.core as pymapdl
from ansys.mapdl.core.mapdl import _MapdlCore
from ansys.mapdl.core.misc import supress_logging

ROUTINE_MAP = {0: 'Begin level',
               17: 'PREP7',
               21: 'SOLUTION',
               31: 'POST1',
               36: 'POST26',
               52: 'AUX2',
               53: 'AUX3',
               62: 'AUX12',
               65: 'AUX15'}

UNITS_MAP = {-1: 'NONE',
             0: 'USER',
             1: 'SI',
             2: 'CGS',
             3: 'BFT',
             4: 'BIN',
             5: 'MKS',
             6: 'MPA',
             7: 'uMKS'}


class Parameters():
    """Collection of MAPDL parameters obtainable from the \*GET command

    Examples
    --------
    Simply list all parameters except for MAPDL MATH parameters

    >>> mapdl.parameters
    ARR                              : ARRAY DIM (3, 1, 1)
    PARM_FLOAT                       : 20.0
    PARM_INT                         : 10.0
    PARM_LONG_STR                    : "stringstringstringstringstringst"
    PARM_STR                         : "string"
    PORT                             : 50052.0

    Get a parameter

    >>> mapdl.parameters['PARM_FLOAT']
    20.0

    Get an array parameter

    >>> mapdl.parameters['ARR']
    array([1., 2., 3.])

    """

    def __init__(self, mapdl):
        if not isinstance(mapdl, _MapdlCore):
            raise TypeError('Must be implemented from MAPDL class')
        self._mapdl_weakref = weakref.ref(mapdl)

    @property
    def _mapdl(self):
        """Return the weakly referenced instance of mapdl"""
        return self._mapdl_weakref()

    def _set_log_level(self, level):
        self._mapdl.set_log_level(level)

    @property
    def _log(self):
        return self._mapdl._log

    @property
    def routine(self) -> str:
        """Current routine string as a string.  For example ``"/PREP7"``

        MAPDL Command: \*GET, ACTIVE, 0, ROUT

        Returns
        -------
        routine : str
            Routine as a string.  One of:

                - ``"Begin level"``
                - ``"PREP7"``
                - ``"SOLUTION"``
                - ``"POST1"``
                - ``"POST26"``
                - ``"AUX2"``
                - ``"AUX3"``
                - ``"AUX12"``
                - ``"AUX15"``

        Examples
        --------
        >>> mapdl.parameters.routine
        'PREP7'
        """
        value = self._mapdl.get_value('ACTIVE', item1='ROUT')
        return ROUTINE_MAP[int(value)]

    @property
    def units(self) -> str:
        """Units specified by /UNITS command.

        Returns
        -------
        units : str
            Active Units.  One of:
                - ``"None"``
                - ``"USER"``
                - ``"SI"``
                - ``"CGS"``
                - ``"BFT"``
                - ``"BIN"``
                - ``"MKS``
                - ``"MPA"``
                - ``"uMKS"``

        Examples
        --------
        >>> mapdl.parameters.units
        'NONE'
        """
        value = self._mapdl.get_value("ACTIVE", item1="UNITS")
        return UNITS_MAP[int(value)]

    @property
    def revision(self) -> float:
        """MAPDL revision version.

        Examples
        --------
        >>> mapdl.parameters.revision
        20.2
        """
        return float(self._mapdl.get_value("ACTIVE", item1="REV"))

    @property
    def platform(self) -> str:
        """The current platform.

        Examples
        --------
        >>> mapdl.parameters.platform
        'LIN'
        """
        return self._mapdl.get_value("ACTIVE", item1="PLATFORM")

    @property
    def csys(self) -> int:
        """Active coordinate system

        Examples
        --------
        >>> mapdl.parameters.csys
        0
        """
        return int(self._mapdl.get_value("ACTIVE", item1="CSYS"))

    @property
    def dsys(self) -> int:
        """Active display coordinate system

        Examples
        --------
        >>> mapdl.parameters.dsys
        0
        """
        return int(self._mapdl.get_value("ACTIVE", item1="DSYS"))

    @property
    def rsys(self) -> int:
        """Active result coordinate system

        Examples
        --------
        >>> mapdl.parameters.rsys
        0
        """
        return int(self._mapdl.get_value("ACTIVE", item1="RSYS"))

    @property
    def esys(self) -> int:
        """Active element coordinate system

        Examples
        --------
        >>> mapdl.parameters.esys
        0
        """
        return int(self._mapdl.get_value("ACTIVE", item1="ESYS"))

    @property
    def section(self) -> int:
        """Active section number

        Examples
        --------
        >>> mapdl.parameters.section
        1
        """
        return int(self._mapdl.get_value("ACTIVE", item1="SECT"))

    @property
    def material(self) -> int:
        """Active material

        Examples
        --------
        >>> mapdl.parameters.material
        1
        """
        return int(self._mapdl.get_value("ACTIVE", item1="MAT"))

    @property
    def real(self) -> int:
        """Active real constant set

        Examples
        --------
        >>> mapdl.parameters.real
        1
        """
        return int(self._mapdl.get_value("ACTIVE", item1="REAL"))

    @property
    def type(self) -> int:
        """Active element type

        Examples
        --------
        >>> mapdl.parameters.type
        1
        """
        return int(self._mapdl.get_value("ACTIVE", item1="type"))

    @property
    @supress_logging
    def _parm(self):
        """Current MAPDL parameters"""
        return interp_star_status(self._mapdl.starstatus())

    def __repr__(self):
        """Return the current parameters in a pretty format"""
        lines = ['MAPDL Parameters',
                 '----------------']
        for key, info in self._parm.items():
            value_str = ''
            if info['type'] == 'ARRAY':
                value_str = 'ARRAY DIM %s' % str(info['shape'])
            elif info['type'] == 'TABLE':
                value_str = 'TABLE DIM %s' % str(info['shape'])
            elif info['type'] == 'CHARACTER':
                value_str = '"%s"' % info['value']
            elif 'value' in info:
                value_str = str(info['value'])
            else:
                continue
            lines.append('%-32s : %s' % (key, value_str))
        return '\n'.join(lines)

    def __getitem__(self, key):
        """Return a parameter"""
        if not isinstance(key, str):
            raise TypeError('Parameter name must be a string')
        key = key.upper()

        parameters = self._parm
        if key not in parameters:
            raise IndexError('%s not a valid parameter_name' % key)

        parm = parameters[key]
        if parm['type'] in ['ARRAY', 'TABLE']:
            try:
                return self._get_parameter_array(key, parm['shape'])
            except ValueError:
                # allow a second attempt
                return self._get_parameter_array(key, parm['shape'])

        return parm['value']

    def __setitem__(self, key, value):
        """Set a parameter"""
        # parameters = self._parm  # check parameter exists
        if isinstance(value, (np.ndarray, list)):
            self._set_parameter_array(key, value)
        else:
            self._set_parameter(key, value)

    @supress_logging
    def _set_parameter(self, name, value):
        """Set a single parameter within MAPDL

        Parameters
        ----------
        name : str
            An alphanumeric name used to identify this parameter.  Name
            may be up to 32 characters, beginning with a letter and
            containing only letters, numbers, and underscores.
            Examples: ``"ABC" "A3X" "TOP_END"``.

        """
        if not isinstance(value, (str, int, float)):
            raise TypeError('``Parameter`` must be either a float, int, or string')

        if not isinstance(name, str):
            raise TypeError('``name`` must be a string')

        if len(name) > 32:
            raise ValueError('Length of ``name`` must be 32 characters or less')

        # delete the parameter if it exists as an array
        parm = self._parm
        if name in parm:
            if parm[name]['type'] == 'ARRAY':
                self._mapdl.starset(name)

        if isinstance(value, str):
            if ' ' in value:
                raise ValueError('Spaces not allowed in strings in MAPDL')
            self._mapdl.starset(name, "'%s'" % value)
        else:
            self._mapdl.starset(name, value)

    @supress_logging
    def _get_parameter_array(self, parm_name, shape):
        """Return an ANSYS array parameter as a numpy.ndarray

        Parameters
        ----------
        parm_name : str
            MAPDL parameter name.

        Returns
        -------
        array : np.ndarray
            Numpy array.
        """
        format_str = '(1F20.12)'
        with self._mapdl.non_interactive:
            self._mapdl.mwrite(parm_name.upper(), label='kji')  # use C ordering
            self._mapdl.run(format_str)

        st = self._mapdl.last_response.rfind(format_str) + len(format_str) + 1
        arr_flat = np.fromstring(self._mapdl.last_response[st:], sep='\n')
        return arr_flat.reshape(shape).squeeze()

    @supress_logging
    def _set_parameter_array(self, name, arr):
        """Load a numpy array or python list directly to MAPDL

        Writes the numpy array to disk and then reads it in within
        MAPDL using \*VREAD.

        Parameters
        ----------
        arr : np.ndarray or List
            Array to send to MAPDL.  Maximum of 3 dimensions.

        name : str
            Name of the array to write to within MAPDL.

        Examples
        --------
        Load a 1D numpy array into MAPDL

        >>> arr = np.array([10, 20, 30])
        >>> mapdl.set_parameter_array(arr, 'MYARR')
        >>> parm, mapdl_arrays = mapdl.load_parameters()
        >>> mapdl_arrays['MYARR']
        array([10., 20., 30.])

        Load a 2D numpy array into MAPDL

        >>> arr = np.random.random((5, 3))
        >>> mapdl.set_parameter_array(arr, 'MYARR')
        >>> parm, mapdl_arrays = mapdl.load_parameters()
        >>> mapdl_arrays['MYARR']
        array([[0.39806635, 0.15060953, 0.3990557 ],
               [0.26837768, 0.02033222, 0.15655861],
               [0.46110226, 0.06381489, 0.20068533],
               [0.20122863, 0.5727896 , 0.85636037],
               [0.68126612, 0.67460878, 0.3678797 ]])

        Load a python list into MAPDL

        >>> mapdl.set_parameter_array([10, -1, 8, 4, 10], 'MYARR')
        >>> parm, mapdl_arrays = mapdl.load_parameters()
        >>> mapdl_arrays['MYARR']
        array([10., -1.,  8.,  4., 10.])

        """
        # type checks
        arr = np.array(arr)
        if not np.issubdtype(arr.dtype, np.number):
            raise TypeError('Only numerical arrays or lists are supported')
        if arr.ndim > 3:
            raise ValueError('MAPDL VREAD only supports a arrays with a'
                             ' maximum of 3 dimensions.')

        if not isinstance(name, str):
            raise TypeError('``name`` must be a string')

        name = name.upper()

        idim, jdim, kdim = arr.shape[0], 0, 0
        if arr.ndim >= 2:
            jdim = arr.shape[1]
        if arr.ndim == 3:
            kdim = arr.shape[2]

        # write array from numpy to disk
        filename = '_tmp.dat'
        self._write_numpy_array(filename, arr)

        self._mapdl.dim(name, imax=idim, jmax=jdim, kmax=kdim)
        with self._mapdl.non_interactive:
            self._mapdl.vread('%s(1, 1),%s,,,IJK, %d, %d, %d' % (name,
                                                                 filename,
                                                                 idim,
                                                                 jdim,
                                                                 kdim))
            self._mapdl.run('(1F20.12)')

    def _write_numpy_array(self, filename, arr):
        """Write a numpy array to disk"""
        if arr.dtype != np.double:
            arr = arr.astype(np.double)

        if self._mapdl._local:
            filename = os.path.join(self._mapdl.directory, filename)
        else:
            filename = os.path.join(tempfile.gettempdir(), filename)
        write_array(filename.encode(), arr.ravel('F'))

        if not self._mapdl._local:
            self._mapdl.upload(filename, progress_bar=False)


def interp_star_status(status):
    """Interprets \*STATUS command output from MAPDL

    Parameters
    ----------
    status : str
        Output from MAPDL *STATUS

    Returns
    -------
    parameters : dict
        Dictionary of parameters.
    """
    parameters = {}
    st = status.find('NAME                              VALUE')
    for line in status[st+80:].splitlines():
        items = line.split()
        if not items:
            continue

        # line will contain either a character, scalar, or array
        name = items[0]
        if len(items) == 2:
            if items[1][-9:] == 'CHARACTER':
                parameters[name] = {'type': 'CHARACTER',
                                    'value': items[1][:-9]}
            # else:
                # log.warning(
        elif len(items) == 3:
            if items[2] == 'SCALAR':
                value = float(items[1])
            else:
                value = items[1]
            parameters[name] = {'type': items[2],
                                'value': value}
        elif len(items) == 5:
            shape = (int(items[2]), int(items[3]), int(items[4]))
            parameters[name] = {'type': items[1],
                                'shape': shape}
    return parameters



import os

from ansys.mapdl.core import __version__, mapdl_functions
from ansys.mapdl.core.misc import is_float

VALID_COMMANDS = dir(mapdl_functions._MapdlCommands)
NON_INTERACTIVE_COMMANDS = ['*CRE', '*VWR']


def convert_script(filename_in, filename_out, loglevel='WARNING', auto_exit=True,
                   line_ending=None, exec_file=None, macros_as_functions=True,
                   use_function_names=True):
    """Converts an ANSYS input file to a python PyMAPDL script.

    Parameters
    ----------
    filename_in : str
        Filename of the ansys input file to read in.

    filename_out : str
        Filename of the python script to write a translation to.

    loglevel : str, optional
        Logging level of the ansys object within the script.

    auto_exit : bool, optional
        Adds a line to the end of the script to exit MAPDL.  Default
        ``True``.

    line_ending : str, optional
        When None, automatically determined by OS being used.

    macros_as_functions : bool, optional
        Attempt to convert MAPDL macros to python functions.

    use_function_names : bool, optional
        Convert MAPDL functions to ansys.mapdl.core.Mapdl class
        methods.  When ``True``, the MAPDL command "K" will be
        converted to ``mapdl.k``.  When ``False``, it will be
        converted to ``mapdl.run('k')``.

    Returns
    -------
    clines : list
        List of lines translated.

    Examples
    --------
    >>> import ansys.mapdl.core as pymapdl
    >>> from ansys.mapdl.core import examples
    >>> clines = pymapdl.convert_script(examples.vmfiles['vm1'], 'vm1.py')
    """
    translator = FileTranslator(loglevel, line_ending, exec_file=exec_file,
                                macros_as_functions=macros_as_functions,
                                use_function_names=use_function_names)
    with open(filename_in) as file_in:
        for line in file_in.readlines():
            translator.translate_line(line)

    if auto_exit:
        translator.write_exit()
    translator.save(filename_out)
    return translator.lines


class FileTranslator():
    obj_name = 'mapdl'
    indent = ''
    non_interactive = False

    def __init__(self, loglevel='INFO', line_ending=None, exec_file=None,
                 macros_as_functions=True, use_function_names=True):
        self._non_interactive_level = 0
        self.lines = []
        self._functions = []
        if line_ending:
            self.line_ending = line_ending
        else:
            self.line_ending = os.linesep
        self.macros_as_functions = macros_as_functions
        self._infunction = False
        self.use_function_names = use_function_names
        self.comment = ''

        self.write_header()
        self.initialize_mapdl_object(loglevel, exec_file)

    def write_header(self):
        hinfo = 'Script generated by ansys-mapdl-core version %s' % __version__
        header = '""" %s """%s' % (hinfo, self.line_ending)
        self.lines.append(header)

    def write_exit(self):
        self.lines.append('%s.exit()%s' % (self.obj_name, self.line_ending))

    def save(self, filename):
        """ Saves lines to file """
        if os.path.isfile(filename):
            os.remove(filename)
        with open(filename, 'w') as f:
            for line in self.lines:
                f.write(line)

    def initialize_mapdl_object(self, loglevel, exec_file):
        """ Initializes ansys object as lines """
        core_module = 'ansys.mapdl.core'
        self.lines.append('from %s%s import launch_mapdl' % (core_module,
                                                             self.line_ending))
        if exec_file:
            exec_file_parameter = '"%s", ' % exec_file
        else:
            exec_file_parameter = ''
        line = '%s = launch_mapdl(%sloglevel="%s")%s' % (self.obj_name,
                                                         exec_file_parameter,
                                                         loglevel,
                                                         self.line_ending)
        self.lines.append(line)

    @property
    def line_ending(self):
        return self._line_ending

    @line_ending.setter
    def line_ending(self, line_ending):
        if line_ending not in ['\n', '\r\n']:
            raise Exception('Line ending must be either "\\n", "\\r\\n"')
        self._line_ending = line_ending

    def translate_line(self, line):
        """Converts a single line from an ANSYS APDL script """
        self.comment = ''
        line = line.strip()
        line = line.replace('"', "'")

        # check if line contains a comment
        if '!' in line:
            if "'!'" in line or '"!"' in line:
                pass
            elif line[0] == '!':  # entire line is a comment
                self.comment = line.replace('!', '').strip()
                self.store_comment()
                return
            else:  # command and in-line comment
                split_line = line.split('!')
                line = split_line[0]
                self.comment = ' '.join(split_line[1:])
                self.comment = self.comment.lstrip()

        if not line:
            return

        if '*END' in line:
            if self.macros_as_functions:
                self.store_empty_line()
                self.store_empty_line()
                self.indent = self.indent[4:]
                self._infunction = False
                self.end_non_interactive()
                return
            else:
                self.store_run_command(line)
                self.end_non_interactive()
                return

        # check for if statement
        if line[:3].upper() == '*IF' or '*IF' in line.upper():
            self.start_non_interactive()
            self.store_run_command(line)
            return

        # check if line ends non-interactive
        if line[0] == '(':
            if not self.non_interactive:
                print('Possible invalid line:\n%s\n' % line +
                      'This line requires a *VWRITE beforehand')
            self.store_run_command(line)
            self.end_non_interactive()
            return
        elif line[:4] == '*USE' and self.macros_as_functions:
            items = line.split(',')
            func_name = items[1].strip()
            if func_name in self._functions:
                args = ', '.join(items[2:])
                call_line = '%s(%s)%s' % (func_name, args, self.line_ending)
                self.lines.append(call_line)
                return

        # check if a line is setting a variable
        items = line.split(',')
        if '=' in items[0]:  # line sets a variable:
            self.store_run_command(line)
            return

        command = items[0].lower().strip()
        parameters = items[1:]
        if not command:
            self.store_empty_line()
            return

        # check valid command
        if command not in VALID_COMMANDS:
            if line[:4] == '*CRE':  # creating a function
                if self.macros_as_functions:
                    self.start_function(items[1].strip())
                    return
                else:
                    self.start_non_interactive()
            elif line[:4] in NON_INTERACTIVE_COMMANDS:
                self.start_non_interactive()
            self.store_run_command(line)
        elif self.use_function_names:
            self.store_command(command, parameters)
        else:
            self.store_run_command(line)

    def start_function(self, func_name):
        self._functions.append(func_name)
        self.store_empty_line()
        self.store_empty_line()
        self._infunction = True
        spacing = ' '*(len(func_name) + 5)
        line = 'def %s(%s,%s' % (func_name, ', '.join(["ARG%d=''" % i for i in range(1, 7)]),
                                 self.line_ending)
        line += '%s%s,%s' % (spacing, ', '.join(["ARG%d=''" % i for i in range(7, 13)]),
                             self.line_ending)
        line += '%s%s):%s' % (spacing, ', '.join(["ARG%d=''" % i for i in range(13, 19)]),
                              self.line_ending)
        self.lines.append(line)
        self.indent = self.indent + '    '

    def store_run_command(self, command):
        """Stores pyansys.ANSYS command that cannot be broken down
        into a function and parameters.
        """
        if self._infunction and 'ARG' in command:
            args = []
            for i in range(1, 19):
                arg = 'ARG%d' % i
                c = 0
                if arg in command:
                    command = command.replace(arg, '{%d:s}' % c)
                    args.append(arg)
                    c += 1

            line = '%s%s.run("%s".format(%s))%s' % (self.indent, self.obj_name, command,
                                                    ', '.join(args), self.line_ending)

        elif self.comment:
            line = '%s%s.run("%s")  # %s%s' % (self.indent, self.obj_name, command,
                                               self.comment, self.line_ending)
        else:
            line = '%s%s.run("%s")%s' % (self.indent, self.obj_name, command,
                                         self.line_ending)

        self.lines.append(line)

    def store_comment(self):
        """Stores a line containing only a comment"""
        line = '%s# %s%s' % (self.indent, self.comment, self.line_ending)
        self.lines.append(line)

    def store_empty_line(self):
        """Stores an empty line"""
        self.lines.append(self.line_ending)

    def store_command(self, function, parameters):
        """Stores a valid pyansys function with parameters"""
        parsed_parameters = []
        for parameter in parameters:
            parameter = parameter.strip()
            if is_float(parameter):
                parsed_parameters.append(parameter)
            elif 'ARG' in parameter and self._infunction:
                parsed_parameters.append('%s' % parameter)
            else:
                parsed_parameters.append('"%s"' % parameter)

        parameter_str = ', '.join(parsed_parameters)
        if self.comment:
            line = '%s%s.%s(%s)  #%s%s' % (self.indent, self.obj_name, function,
                                           parameter_str, self.comment, self.line_ending)
        else:
            line = '%s%s.%s(%s)%s' % (self.indent, self.obj_name, function,
                                      parameter_str, self.line_ending)

        self.lines.append(line)

    def start_non_interactive(self):
        self._non_interactive_level += 1
        if self.non_interactive:
            return
        line = f'{self.indent}with {self.obj_name}.non_interactive:{self.line_ending}'
        self.lines.append(line)
        self.non_interactive = True
        self.indent = self.indent + '    '

    def end_non_interactive(self):
        self._non_interactive_level -= 1
        if self._non_interactive_level == 0:
            self.non_interactive = False
            self.indent = self.indent[4:]
